package org.nusco.narjillos.creature.body;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import org.nusco.narjillos.creature.body.physics.RotationalForceField;
import org.nusco.narjillos.creature.body.physics.TranslationalForceField;
import org.nusco.narjillos.shared.physics.Segment;
import org.nusco.narjillos.shared.physics.Vector;
import org.nusco.narjillos.shared.physics.ZeroVectorException;

/**
 * The physical body of a Narjillo, with all its organs and their position in space.
 * 
 * This class contains the all-important Body.tick() method. Look at its comments
 * for details.
 */
public class Body {

	private final Organ head;
	private final double mass;
	private transient List<BodyPart> bodyParts;
	
	public Body(Head head) {
		this.head = head;
		this.mass = calculateTotalMass();
	}

	public Head getHead() {
		return (Head) head;
	}

	public double getMass() {
		return mass;
	}

	public List<BodyPart> getBodyParts() {
		if (bodyParts == null) {
			bodyParts = new ArrayList<>();
			addWithChildren(bodyParts, getHead());
		}
		return bodyParts;
	}

	private void addWithChildren(List<BodyPart> result, Organ organ) {
		// children first
		for (Organ child : organ.getChildren())
			addWithChildren(result, child);
		result.add(organ);
	}

	public void teleportTo(Vector position) {
		final int northDirection = 90;
		getHead().moveTo(position, northDirection);
	}

	public Vector getStartPoint() {
		return getHead().getStartPoint();
	}

	public Vector calculateCenterOfMass() {
		if (getMass() <= 0)
			return getStartPoint();

		// do it in one swoop instead of creating a lot of
		// intermediate vectors

		List<BodyPart> organs = getBodyParts();
		Vector[] weightedCentersOfMass = new Vector[organs.size()];
		Iterator<BodyPart> iterator = getBodyParts().iterator();
		for (int i = 0; i < weightedCentersOfMass.length; i++) {
			BodyPart organ = iterator.next();
			weightedCentersOfMass[i] = organ.getCenterOfMass().by(organ.getMass());
		}
		
		double totalX = 0;
		double totalY = 0;
		for (int i = 0; i < weightedCentersOfMass.length; i++) {
			totalX += weightedCentersOfMass[i].x;
			totalY += weightedCentersOfMass[i].y;
		}
		
		return Vector.cartesian(totalX / getMass(), totalY / getMass());
	}

	/**
	 * Take a target direction. Change the body's geometry based on the target
	 * direction. Move the body. Return the energy spent on the entire operation.
	 */
	public double tick(Vector targetDirection) {
		// Before any movement, store away the current body positions
		// and center of mass. These will come useful later.
		Vector initialCenterOfMass = calculateCenterOfMass();
		Map<BodyPart, Segment> initialBodyPartPositions = calculateBodyPartPositions();
		
		// This first step happens as if the body where in a vacuum.
		// The organs in the body remodel their own geometry based on the
		// target's direction. They don't "think" were to go - they just
		// changes their positions *somehow*. Natural selection will favor
		// movements that result in getting closer to the target.
		tick_step1_updateAngles(targetDirection);
		
		// Changing the angles in the body results in a rotational force.
		// Rotate the body to match the force. In other words, keep its moment
		// of inertia equal to zero.
		double rotationEnergy = tick_step2_rotate(initialCenterOfMass, initialBodyPartPositions);

		// The previous updates moved the center of mass. Remember, we're
		// in a vacuum - so the center of mass shouldn't move. Let's put it
		// back to its original position.
		tick_step3_recenter(initialCenterOfMass);

		// Now we can finally move out of the "vacuum" reference system.
		// All the movements from the previous steps result in a different
		// body position in space, and this different position generates
		// translational forces. We can update the body position based on
		// these translations.
		double translationEnergy = tick_step4_translate(initialBodyPartPositions, initialCenterOfMass);
		
		return (rotationEnergy + translationEnergy) * getMetabolicRate();
	}

	private void tick_step1_updateAngles(Vector targetDirection) {
		double angleToTarget = getAngleTo(targetDirection);
		getHead().recursivelyUpdateAngleToParent(0, angleToTarget);
	}

	private double getAngleTo(Vector direction) {
		if (direction.equals(Vector.ZERO))
			return 0;
		try {
			return getHead().getMainAxis().getAngleWith(direction);
		} catch (ZeroVectorException e) {
			throw new RuntimeException(e); // should never happen
		}
	}

	private double tick_step2_rotate(Vector centerOfMass, Map<BodyPart, Segment> initialPositions) {
		RotationalForceField forceField = new RotationalForceField(getMass(), calculateRadius(centerOfMass), centerOfMass);
		for (BodyPart bodyPart : bodyParts)
			forceField.registerMovement(initialPositions.get(bodyPart), bodyPart.getPositionInSpace(), bodyPart.getMass());
		getHead().moveBy(Vector.ZERO, forceField.getRotation());
		return forceField.getEnergy();
	}

	private void tick_step3_recenter(Vector centerOfMassBeforeReshaping) {
		Vector centerOfMassAfterUpdatingAngles = calculateCenterOfMass();
		Vector centerOfMassOffset = centerOfMassBeforeReshaping.minus(centerOfMassAfterUpdatingAngles);
		getHead().moveBy(centerOfMassOffset, 0);
	}

	private double tick_step4_translate(Map<BodyPart, Segment> initialPositions, Vector centerOfMass) {
		TranslationalForceField forceField = new TranslationalForceField(getMass());
		for (BodyPart bodyPart : bodyParts)
			forceField.registerMovement(initialPositions.get(bodyPart), bodyPart.getPositionInSpace(), bodyPart.getMass());
		getHead().moveBy(forceField.getTranslation(), 0);
		return forceField.getEnergy();
	}

	private Map<BodyPart, Segment> calculateBodyPartPositions() {
		Map<BodyPart, Segment> result = new LinkedHashMap<>();
		for (BodyPart bodyPart : getBodyParts())
			result.put(bodyPart, bodyPart.getPositionInSpace());
		return result;
	}

	double calculateRadius(Vector centerOfMass) {
		final double MIN_RADIUS = 1;
		double result = MIN_RADIUS;
		for (BodyPart bodyPart : getBodyParts()) {
			double startPointDistance = bodyPart.getStartPoint().minus(centerOfMass).getLength();
			double endPointDistance = bodyPart.getEndPoint().minus(centerOfMass).getLength();
			double distance = Math.max(startPointDistance, endPointDistance);
			if (distance > result)
				result = distance;
		}
		return result;
	}

	private double getMetabolicRate() {
		return getHead().getMetabolicRate();
	}

	public double getPercentEnergyToChildren() {
		return getHead().getPercentEnergyToChildren();
	}

	private double calculateTotalMass() {
		double result = 0;
		List<BodyPart> allOrgans = getBodyParts();
		for (BodyPart organ : allOrgans)
			result += organ.getMass();
		return result;
	}
}
